# Copyright (C) 2019 
#
# The source code in this file is released under the GPLv3.
# Go to https://www.gnu.org/licenses/gpl.txt for the full license details.
#
import upygame as pygame
import urandom as random
import umachine

import res
import maps

# Set palette
pygame.display.set_palette_16bit([
    0, 6438, 18917, 10825, 47398, 688, 41764, 17475,
    0xdc9c, 13598, 60486, 40179, 42596, 46845, 63245, 65535
]);

tile_w = 8
tile_h = 8
tiles = {
    "c": pygame.surface.Surface(tile_w, tile_h, res.pixels_crate),
    "w": pygame.surface.Surface(tile_w, tile_h, res.pixels_wall),
    ".": pygame.surface.Surface(tile_w, tile_h, res.pixels_floor),
    "e": pygame.surface.Surface(tile_w, tile_h, res.pixels_exit),
    "s": pygame.surface.Surface(tile_w, tile_h, res.pixels_start),
    "b": pygame.surface.Surface(tile_w, tile_h, res.pixels_btn_off),
    "B": pygame.surface.Surface(tile_w, tile_h, res.pixels_btn_on),
    "o": pygame.surface.Surface(tile_w, tile_h, res.pixels_btndoor_closed),
    "O": pygame.surface.Surface(tile_w, tile_h, res.pixels_btndoor_open),
    "d": pygame.surface.Surface(tile_w, tile_h, res.pixels_door_closed),
    "D": pygame.surface.Surface(tile_w, tile_h, res.pixels_door_open),
    "k": pygame.surface.Surface(tile_w, tile_h, res.pixels_key),
    "E": pygame.surface.Surface(tile_w, tile_h, res.pixels_cat),
}

GO_START = 0
GO_LEVELS = 1
GO_INTRO = 2
GO_OUTRO = 3



class TileSprite(object):
    def __init__(self, images, row=0, column=0):
        self.cur_image = images[0]
        self.images = images
        self.set_coord(row, column)
        self.vx = 0
        self.vy = 0
        self.keep_moving = False
        self.activated = False

    def set_coord(self, row, column):
        self.row = row
        self.column = column
        self.x = column * tile_w
        self.y = row * tile_h

    def update(self):
        if self.vx:
           self.x += self.vx
           if self.x % tile_w == 0:
              self.column = self.x // tile_w
              if self.keep_moving:
                 self.vx = 0
                 self.keep_moving = False

        if self.vy:
           self.y += self.vy
           
           if self.y % tile_h == 0:
               self.row = self.y // tile_h
               if self.keep_moving:
                  self.vy = 0
                  self.keep_moving = False

    def move(self, vx=0, vy=0):
        self.vx = vx
        self.vy = vy
        self.keep_moving = True

    def stop(self):
        self.keep_moving = False # stop on next tile coord

class Key(TileSprite): 
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["k"],), row, column)

class Crate(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["c"],), row, column)

class Button(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["b"],tiles["B"],), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True

class ButtonDoor(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["o"],tiles["O"]), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True

class Door(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["d"],tiles["D"]), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True
        # yuck. duplicate code...

player_images = (
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_down),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_up),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_left),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_right),
    # TODO: Up, down, left, right + anim step
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_down2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_up2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_left2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_right2),
)


class Player(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, player_images, row, column)
        self.keys = 0
        self.pushing_crate = None
        self.anim_timer = 0

    def move(self, vx=0, vy=0):
        TileSprite.move(self, vx, vy)

        if self.vx > 0:
            self.cur_image = self.images[3]
        elif self.vx < 0:
            self.cur_image = self.images[2]
        elif self.vy > 0:
            self.cur_image = self.images[0]
        elif self.vy < 0:
            self.cur_image = self.images[1]
        elif self.pushing_crate:
            self.pushing_crate.stop()
            self.pushing_crate = None

class Map(object):
    blocking = " ws"

    def __init__(self, w, h, map_data):
        self.map_data = map_data
        self.width = w
        self.height = h
        self.offset_x = (110 - self.width * tile_w) // 2
        self.offset_y = (88 - self.height * tile_h) // 2

    def get_tile_pos(self, tile, start=0):
        idx = self.map_data.index(tile, start)
        row = idx // self.width 
        col = idx % self.width
        return row, col

    def draw(self, surface):
        # bottom layer
        for row in range(self.height):
            for col in range(self.width):
                tile = self.map_data[(row * self.width) + col]
                if tile in "ck":
                    tile = "."
                if tile in ".wesE":
                    surface.blit(tiles[tile], col*tile_w + self.offset_x, row*tile_h + self.offset_y)
                
    def is_blocking(self, col, row):
        idx = (row * self.width) + col
        return self.map_data[idx] in self.blocking
        #return False


class GameState:
    screen = None
    u_pressed = False
    d_pressed = False
    l_pressed = False
    r_pressed = False
    a_pressed = False
    b_pressed = False
    exit_status = None
    
    def __init__(self, screen):
        self.screen = screen

    def check_events(self):
        eventtype = pygame.event.poll()
        if eventtype == pygame.NOEVENT:
            pass
        elif eventtype.type == pygame.KEYDOWN:
            if eventtype.key == pygame.K_RIGHT:
                self.r_pressed = True
            elif eventtype.key == pygame.K_LEFT:
                self.l_pressed = True
            elif eventtype.key == pygame.K_UP:
                self.u_pressed = True
            elif eventtype.key == pygame.K_DOWN:
                self.d_pressed = True
            elif eventtype.key == pygame.BUT_A:
                self.a_pressed = True
            elif eventtype.key == pygame.BUT_B:
                self.b_pressed = True
        elif eventtype.type == pygame.KEYUP:
            if eventtype.key == pygame.K_RIGHT:
                self.r_pressed = False
            elif eventtype.key == pygame.K_LEFT:
                self.l_pressed = False
            elif eventtype.key == pygame.K_UP:
                self.u_pressed = False
            elif eventtype.key == pygame.K_DOWN:
                self.d_pressed = False
            elif eventtype.key == pygame.BUT_A:
                self.a_pressed = False
            elif eventtype.key == pygame.BUT_B:
                self.b_pressed = False
            
    
    def update_logic(self):
        pass

    def update_screen(self):
        pygame.display.flip()
    
    def mainloop(self):
        running = True
        while running:
            self.check_events()
            self.update_logic()
            self.update_screen()
            if self.exit_status is not None:
                break
        return self.exit_status

class StartScreen(GameState):
    blinker = 0
    def update_screen(self):
        umachine.draw_text(31, 20, "The Villainy", 15)
        umachine.draw_text(31, 26, "      of    ", 15)
        umachine.draw_text(31, 32, "Cat Food Inc.", 15)
        
        self.blinker += 1
        if self.blinker > 40:
            umachine.draw_text(25, 62, "Press A to Start", 15)
        if self.blinker > 80:
            self.blinker = 0
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed or self.b_pressed:
            self.exit_status = 1

intro_guy = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_stand_big)
intro_johnny = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_johnny_stand_big)
intro_evilguy = pygame.surface.Surface(res.evilguy_w, res.evilguy_h, res.pixels_evilguy_stand_big)
intro_building = pygame.surface.Surface(16, 32, res.buildingPixels);
class IntroScreen(GameState):
    def update_screen(self):
        umachine.draw_text(10, 10, "Cat Food Inc. catnapped", 15)
        umachine.draw_text(10, 16, "Nanji, Guy's cat. ", 15)
        
        self.screen.blit(intro_guy, 10, 30)
        self.screen.blit(intro_evilguy, 60, 30)
        self.screen.blit(intro_johnny, 70, 30)
        self.screen.blit(intro_building, 94, 15)
        
        umachine.draw_text(10, 60, "Help Guy rescue Nanji", 15)
        umachine.draw_text(10, 66, "and escape from the", 15)
        umachine.draw_text(10, 72, "Cat Food Inc. building.", 15)
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed or self.b_pressed:
            self.exit_status = 1


outro_guy = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_stand_cat)
class OutroScreen(GameState):
    def update_screen(self):
        self.screen.blit(outro_guy, 50, 40)
        umachine.draw_text(25, 70, "To be continued...", 15)
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed or self.b_pressed:
            self.exit_status = 1

class MapState(GameState):

    def __init__(self, screen, map_w, map_h, map_data):
        self.cur_map = Map(map_w, map_h, map_data)
        self.exit_coord = None
        self.player = Player()
        self.sprites = []
        self.crates = []
        self.doors = []
        self.keys = []
        self.button = None
        self.btn_door = None

        # find player, crates, doors, buttons
        idx = 0
        for tile in self.cur_map.map_data:
            row = idx // self.cur_map.width
            column = idx % self.cur_map.width
            if tile == "e" or tile == "E":
                self.exit_coord = (column, row)
            elif tile == "c":
                self.crates.append(Crate(row, column))
            elif tile == "s":
                self.player.set_coord(row + 1, column)
            elif tile == "o":
                self.btn_door = ButtonDoor(row, column)
                self.sprites.append(self.btn_door)
            elif tile == "b":
                self.button = Button(row, column)
                self.sprites.append(self.button)
            elif tile == "d":
                self.doors.append(Door(row, column))
            elif tile == "k":
                self.keys.append(Key(row, column))
            idx += 1

        self.sprites.extend(self.doors)
        self.sprites.extend(self.keys)
        self.sprites.extend(self.crates)
        self.sprites.append(self.player) 
        GameState.__init__(self, screen)

    def check_events(self):
        GameState.check_events(self)
        if not self.player.keep_moving:
            if self.u_pressed:
                self.player.move(vy=-1)
            elif self.d_pressed:
                self.player.move(vy=1)
            elif self.l_pressed:
                self.player.move(vx=-1)
            elif self.r_pressed:
                self.player.move(vx=1)
            else: 
                self.player.stop()
        if self.b_pressed:
            self.exit_status = 2

    def update_screen(self):
        self.cur_map.draw(self.screen)
        for s in self.sprites:
            self.screen.blit(s.cur_image, s.x + self.cur_map.offset_x, s.y + self.cur_map.offset_y)
        
        # draw number of keys
        self.screen.blit(tiles["k"], 0, 0)
        umachine.draw_text(8, 3, "x" + str(self.player.keys), 15)
        GameState.update_screen(self)

    def contains_crate(self, column, row):
        for c in self.crates:
            if c.column == column and row == c.row:
                return c
        return None

    def contains_door(self, column, row):
        for d in self.doors:
            if d.column == column and d.row == row:
                return d
        return None
    
    def contains_key(self, column, row):
        for k in self.keys:
            if k.column == column and k.row == row:
                return k
        return None
    
    def win(self):
        self.exit_status = 0

    def update_logic(self):
        # monstrosity... 
        for s in self.sprites:
            s.update()
            
            if s.vx or s.vy:
                new_row = s.row + s.vy
                new_col = s.column + s.vx
                stop_moving = False

                # check "collisions"
                if self.cur_map.is_blocking(new_col, new_row):
                    stop_moving = True
                elif self.button.row == new_row and self.button.column == new_col:
                    stop_moving = True
                    if not self.button.activated:
                        self.button.activate()
                        self.btn_door.activate()
                elif not self.btn_door.activated and new_row == self.btn_door.row and new_col == self.btn_door.column:
                    stop_moving = True

                # what a mess.. again...
                if not stop_moving:
                    door = self.contains_door(new_col, new_row)
                    if door and not door.activated:
                        stop_moving = True
                        if s is self.player and self.player.keys > 0:
                            door.activate()
                            self.player.keys -= 1
                    if not door:
                        crate = self.contains_crate(new_col, new_row)
                        if crate:
                            if s is self.player:
                                # check if crate moveable, push crate
                                #stop_moving = True 
                                #self.player.pushing_crate = crate
                                next_next_row = new_row + s.vy
                                next_next_col = new_col + s.vx
                                if self.cur_map.is_blocking(next_next_col, next_next_row) or \
                                    self.contains_door(next_next_col, next_next_row) or \
                                    self.contains_crate(next_next_col, next_next_row) or \
                                    (next_next_row == self.btn_door.row and next_next_col == self.btn_door.column) or \
                                    (next_next_row == self.btn_door.row and next_next_col == self.btn_door.column):
                                    stop_moving = True
                                else:
                                    crate.move(self.player.vx, self.player.vy)
                                    self.player.pushing_crate = crate
                                    
                            else:
                                # crates cannot push other crates!
                                stop_moving = True 
                                self.player.stop()

                if stop_moving:
                    s.set_coord(s.row, s.column)
                    s.vx = 0
                    s.vy = 0
                    s.stop()
                    
                # moving objects destroy keys .. only player picks them up
                coll_key = self.contains_key(s.column, s.row)
                if coll_key:
                    self.sprites.remove(coll_key)
                    self.keys.remove(coll_key) # this is why spritegroups are so nice, you know...
                    if s is self.player:
                        self.player.keys += 1
            
            # oh, yes. updated the Sprite, pls
            s.update()

        if self.exit_coord == (self.player.column, self.player.row):
            self.win()


def level_loop(screen):
    level_idx = 0
    while level_idx < len(maps.levels):
        cur_level = maps.levels[level_idx]
        state = MapState(screen, cur_level.width, cur_level.height, cur_level.map_data)
        exit_status = state.mainloop()
        if exit_status == 0:
            level_idx += 1  # load next level
        elif exit_status == 1:
            break # goto start menu
        # else repeat level again
    if level_idx == len(maps.levels):
        return 3
    return 0
    
# start game
def main():
    pygame.display.init()
    screen = pygame.display.set_mode()
    exit_status = GO_START
    # state engine loop
    while True:
        if exit_status == GO_OUTRO:
            state = OutroScreen(screen)
            state.mainloop()
            exit_status = 0
        elif exit_status == GO_LEVELS:
            exit_status = level_loop(screen)
        elif exit_status == GO_INTRO:
            state = IntroScreen(screen)
            state.mainloop()
            exit_status = GO_LEVELS
        else:
            state = StartScreen(screen)
            state.mainloop()
            exit_status = GO_INTRO

main()